/************************************************************************tt
*                       - INFOFLEX System -
* This source code is CONFIDENTIAL and PROPRIETARY to CUSTOMWARE COMPUTING,
* INC. (CWC). The information contained herein may not be used, copied or
* disclosed in whole or in part except as permitted by written agreement
* signed by an officer of CWC.
*
* COPYRIGHT (c) 1986 CUSTOMWARE COMPUTING, INC., BURLINGAME, CALIFORNIA.
* ALL RIGHTS RESERVED.
*
* Version: %V%  %R%
* Date: %D%  %T%
*************************************************************************/

/*

DESIGN:

     I.  Layout of structure file.

         Each resident structure (SCRHEAD, DBHEAD, SCRFIELD, DBFIELD, DBINDEX)
         will be contiguous arrays, as before.

            DBHEAD
              :
            SCRHEAD
              :
            INSHEAD
              :
            DBFIELD
              :
            SCRFIELD
              :
            INSFIELD (resident)
              :
            DBINDEX
              :
            RESIDENT TEXT
            SYMBOL TEXT  - one text array for entire program. Includes
                             Symbolic Names for all DBFIELD, DBINDDEX,
                             SCRFIELD, RPTFIELD names.
                             These need not be resident  unless
                             user refers to these names in "C" programs.
                             Therefore, these will be loaded upon
                             request of user.
            TEMPORARY TEXT (intermediate file durring preprocessing)



         At the end of these structures will be transient information:

              TRANSIENT EXTENSION - Possibilities are MENUTRANS, SCRTRANS,
                                    RPTTRANS.
              TRANSIENT INSTRUCTIONS - Local instruction routines.
                                       Store test in TRANSIENT TEXT (below).
              TRANSIENT TEXT      - text array for miscellaneous text.
                                    Example would be screen Comments.
              TRANSIENT LITERALS  - text array for display literals
                                    Literal are stored as 2 shorts for
                                    Row and Column followed by the literal
                                    string.
                :
                :    ---> repeat above 4 for header types:
                          MENUHEAD, SCRHEAD, RPTHEAD


                     DBHEAD and INSHEAD do NOT need TRANSIENT Information.




     II.  Preprocessor

          PASS1:
           - Load all Database information first so I can store index
             numbers for database references.
             You will need to Preprocess xxx.cw file for Table Section only.
             Extract DBFIELD and DBINDEX information from CISAM-SQL
             WRITE  out as:
                 COUNTS --->
                 DBFIELD ---> DBFLD
                 DBINDEX ---> DBIDX
                 SYMBOL --->  SYMBOL
                 RES TEXT ---> RESTEXT
             Reload these files into memory before going on to next step.

          PASS2:
           - Preprocess all other Sections of xxxx.cw files.
             Keep xxxHEAD and RESIDENT TEXT in memory
             and write out xxxFIELD structures
                 SCRFIELD  ---> SCRFLD

           - WRITE out TRANSIENT information TEMP1.

           - WRITE out all RESIDENT information to TEMP2 file.
           - Concatenate TEMP2 and TEMP1 files.

          PASS2:
           - read in structures like running program would.
           - do linking
    III.  Program Initialization

         - Allocate space for HDHEAD
         - Read HDHEAD structure.
         - Allocate space for RESIDENT Information based on
           hdRessize (see HDHEAD)
           RESIDENT information includes all xxxHEAD, xxxINDEX,
           and xxxFIELD structures and a test array.
         - Allocate data buffers based on hdBufsize
         - Allocate maximum space for TRANSIENT Information based on
           hdTranssize (see HDHEAD)
         - Allocate space for SYMBOL Information based on
           hdSymsize (see HDHEAD). This may be optional based on
           flag settings (see HDHEAD).
         - Read all RESIDENT Information by
           reading the next hdRessize bytes (see HDHEAD).
              (DBHEAD, SCRHEAD, INSHEAD)
              (DBFIELD, DBINDEX)
              (SCRFIELD)
              (INSFIELD)

         - Set global pointers for RESIDENT structures.
         - Allocate data buffers for screen and database fields.
         - Convert text array offsets to true pointers ???
           For RESIDENT and SYMBOL text arrays.


*/

#include "fxisam.h"


typedef struct {
    long    trOffset;    /*  offset position of TRANSIENT information  */
    short   trSize;      /*  Total size in bytes for all TRANSIENT Information */
    short   trStrnum;   /*  Number of SCRTRANS structures  */
    short   trStrsize;   /*  Number of SCRTRANS structures  */
    char   *trStrptr;    /*  Pointer to SCRTRANS structures  */
    short   trIfnum;    /*  Number of INSFIELD structures  */
    short   trIfsize;   /*  Number of INSFIELD structures  */
    char   *trIfptr;    /*  Pointer to INSFIELD structures  */
    short   trTxtsize;   /*  Number of bytes in TRANSIENT TEXT ARRAY */
    char   *trTxtptr;    /*  Pointer to TRANSIENT TEXT ARRAY */
    short   trLitnum;    /*  Number of TRANSIENT LITERAL ARRAY */
    short   trLitsize;   /*  Number of bytes in TRANSIENT LITERAL ARRAY */
    char   *trLitptr;    /*  Pointer to TRANSIENT LITERAL ARRAY */
   } TRANSIENT;


typedef struct {
    char   *hName;     /* ---> offset into RESIDENT text array
                          screen name */
    char   *hBuffer;   /* length of screen buffer */
    short   hLength;   /* length of screen buffer */
    short   hIndex;  /* index into array of structs of type FIELD */
    short   hNum;    /* number of fields composing the screen */
    } HEAD;

typedef struct {
    char    *fName;    /* ---> field name - offset into SYMBOL text array*/
    short    fType;    /* field data type (CHARTYPE, INTTYPE, etc.) */
    short    fState;   /* field state - updated by programmer */
    short    fBytes;   /* field width in bytes */
    short    fOffs;  /* field offset within a reserved i/o record buffer */
    short    fIndex;   /*  */
    short    fOther;   /*  */
   } FIELD;



/***********************************************************************
*                                                                      *
*         D A T A B A S E      S T R U C T U R E S                     *
*                                                                      *
************************************************************************/


#ifndef ALL_TYPES
struct dbview
    {
    char *vwname;	/* name of attribute	*/
    int vwstart;	/* offset in view	*/
    int vwtype;		/* type of attribute	*/
    int vwlen;		/* length in bytes	*/
    int vwoffs;		/* Offset of source buffer*/
    };
#endif


typedef struct
{
    char    *dfName;    /* ---> field name - offset into SYMBOL text array*/
    short    dfType;    /* field data type (CHARTYPE, INTTYPE, etc.) */
    short    dfState;  /* field state - updated by programmer */
    short    dfBytes;   /* field width in bytes */
    short    dfOffs;    /* field offset within a reserved i/o record buffer */
    short    dfDhindex; /* index into array of structs of type DBHEAD */
    short    dfDecimal; 
} DBFIELD;



typedef struct
{
    char   *diName;     /* ---> index name - offset into SYMBOL text array */
    short   diType;     /* ---> (U)nique vs. (D)ups */
    short   diDhindex;  /* index into array of structs of type DBHEAD */
    short   diLength;   /* index into array of structs of type DBHEAD */
    short   diDfnum;    /* number of key parts within index     */
    short   diDfindex[ NPARTS]; /* negative for descending order  */
} DBINDEX;

typedef struct 
{
     char daName[20];
     int  daFldesc;
     DBINDEX *daDbindex;
} DBALTINDEX;

typedef struct
{
    char    *dhName;     /* ---> offset into RESIDENT text array
                           file name */
    char    *dhPath;     /* ---> offset into RESIDENT text array  */
    short    dhLength;   /* length of each record              */
    short    dhFldesc;   /* file # assigned by C-ISAM's when file opened.*/
                         /* Value is -1 when file is closed.        */
    short    dhFlmode;   /* access mode of file when opened */
    short    dhState;    /* state of file */

    DBFIELD *dhDbfield;  /* index into array of structs of type DBFIELD */
    short    dhDfnum;    /* number of fields per record */
    short    dhDfserial; /* index into array of structs of type DBFIELD
                            for SERIAL field if it exists */
    DBINDEX *dhDbindex;  /* index into array of structs of type DBINDEX */
    short    dhDinum;    /* number of elements in abovementioned array */
    char    *dhDicurrent; /* Currently active index set at runtime - struct type DBINDEX*/
    char    *dhBuffer;   /* buffer where screen data should be stored.    */
    struct dbview  *dhView;     /* pointer to dbview structure */
    char    *dhVwstruct; /* pointer to structure where dbview fields will be stored.*/
    short    dhVwnum;    /* number of fields in dbview structure.    */
    short    dhTabid;    /* number of fields in dbview structure.    */
    DBALTINDEX *dhDbaltindex;
} DBHEAD;



/***********************************************************************
*                                                                      *
*              S C R E E N     S T R U C T U R E S                     *
*                                                                      *
************************************************************************/

typedef struct
{
    short kk_key ;      /* key value  */
    /* int (*kk_func)(); */   /* function called by key */
    char *kk_label;      /* labels  */
} KEYTABLE;




typedef struct
{
    char    *sfName;   /* ---> screen field name as found in brackets */
                       /* offset into SYMBOL text array */
    short    sfType;   /* Field Types  (see dclcon.h) */
    short    sfState;  /* field state - updated by programmer */
    short    sfBytes; /* display length in ASCII format */
    short    sfOffs;   /* offset into screen data buffer */

    short    sfShindex;/* index into array of struct of type SCRHEAD */
    short    sfDfindex;/* index into the array of struct of type DBFIELD */
                       /* which corresponds to this screen field */
                       /* -1 value means not a database field */

} SCRFIELD;


typedef struct
{
    short  stLength; /* Visual length of screen field */
    short  stDec;    /* Number of places to the right of decimal point */
    short  stAttr;   /* field attribute flags (reverse, autonext)*/
    short  stAttr2;  /* field format flags (i.e. comma, phone ... ) */
    short  stAttr3;  /* field format flags (i.e. comma, phone ... ) */
    short  stFormat; /* field format flags (i.e. comma, phone ... ) */
    short  stRow;    /* Physical Row position on the screen */
    short  stCol;    /* Physical Column position on the screen */
    short  stComment; /* ---> offset into TRANSIENT text array */
    short  stDefault; /* ---> offset into TRANSIENT text array  */
    short  stInclude; /* ---> offset into TRANSIENT text array  */
    short  stFmtfield;/* index into array of structure INSHEAD  */
    short  stZoom;    /* ---> offset into TRANSIENT text array  */
    short  stSearchby;/* Index into array of struct of type DBINDEX */
    short  stLookup;  /* index into array of structure DBINDEX */
    short  stLookfunc; /* index into array of structure INSHEAD */
    short  stLookinput;/* index into array of structure INSHEAD */
    short  stLookinfunc; /* index into array of structure INSHEAD */
    short  stHelp;    /* index into array of structure DBINDEX */
    short  stHmenu;   /* Menu chainword - stored in TRANSIENT TEXT */
    short  stHfld1;   /* index into array of structure DBFIELD */
    short  stHfld2;   /* index into array of structure DBFIELD */
    short  stHfld3;   /* index into array of structure DBFIELD */
    short  stTotal;   /* index into array of structure SCRFIELD */
    short  stBefunc;  /* index into array of structure INSHEAD */
    short  stWfunc;   /* index into array of structure INSHEAD */
    short  stAfunc;   /* index into array of structure INSHEAD */
    short  stFfunc;   /* index into array of structure INSHEAD */
} SCRTRANS;


typedef struct
{
    char   *shName;     /* ---> offset into RESIDENT text array
                          screen name */
    short   shType;     /* SCREEN/REPORT/PROCESS text array
                          screen name */
    short   shParent;
    short   shLink;
    short   shLength;   /* length of screen buffer */

    short   shToprow;   /* starting row position of screen */
    short   shTopcol;   /* starting column position of screen */
    short   shRows;     /* number of rows in screen */
    short   shCols;     /* number of columns in screen */

    short   shAshindex; /* index into array of structs of type SCRFIELD */
    short   shArepeat;  /* number of rows in the array  */
    short   shRownum ;  /* Active array row number */

    int     shCurrow ;  /* row offset added to row - updated by user */
    int     shCurcol;   /* column offset added to column - updated by user */

    short   shXBsection; /* userexit -  Before entering screen/report */
    short   shXAsection; /* userexit -  After exiting screen/report */
    short   shXBsubsection;/* userexit -  Upon entering flex program */
    short   shXAsubsection;/* userexit -  Upon exiting flex program */
    short   shXBrow;    /* userexit -  Before positioning onto row */
    short   shXArow;    /* userexit -  After leaving row  */
    short   shXBsave;   /* userexit -  Before saving row */
    short   shXDsave;   /* userexit -  Before saving row */
    short   shXAsave;   /* userexit -  After saving row */
    short   shXBdelete; /* userexit -  Before deleting row */
    short   shXDdelete; /* userexit -  Before deleting row */
    short   shXAdelete; /* userexit -  After deleting row */
    short   shXBdisplay; /* userexit -  Before displaying array line buffer */
    short   shXselect;  /* userexit -  Select record for array line */
    short   shXenterkey;  /* userexit -  Upon pressing ZOOMKEY function key*/
    short   shXescapekey;  /* userexit -  Upon pressing ZOOMKEY function key*/
    short   shXuparrowkey;  /* userexit - Upon pressing UPARROW function key*/
    short   shXdownarrowkey;/* userexit - Upon pressing DOWNARROW function key*/
    short   shXbacktabkey;  /* userexit - Upon pressing UPARROW function key*/
    short   shXsavekey;  /* userexit -  Upon pressing ZOOMKEY function key*/
    short   shXdupkey;  /* userexit -  Upon pressing ZOOMKEY function key*/
    short   shXjumpkey;  /* userexit -  Upon pressing JUMPKEY function key*/
    short   shXmodekey;  /* userexit -  Upon pressing MODEKEY function key*/
    short   shXBsearchkey;/* userexit -  Upon pressing SEARCHKEY function key*/
    short   shXAsearchkey;/* userexit -  Upon pressing SEARCHKEY function key*/
    short   shXfindkey;  /* userexit -  Upon pressing FINDKEY function key*/
    short   shXprevkey;  /* userexit -  Upon pressing PREVKEY function key*/
    short   shXnextkey;  /* userexit -  Upon pressing NEXTKEY function key*/
    short   shXfirstkey; /* userexit -  Upon pressing FIRSTKEY function key*/
    short   shXlastkey;  /* userexit -  Upon pressing LASTKEY function key*/
    short   shXprintkey; /* userexit -  Upon pressing PRINTKEY functin key*/
    short   shXdelkey;  /* userexit -  Upon pressing ZOOMKEY function key*/
    short   shXzoomkey;  /* userexit -  Upon pressing ZOOMKEY function key*/
    short   shXhelpkey;  /* userexit -  Upon pressing HELPKEY function key*/
    short   shXquerykey; /* userexit -  Upon pressing SKIPKEY function key*/
    short   shXskipkey;  /* userexit -  Upon pressing SKIPKEY function key*/
    short   shXu1key;    /* userexit -  Upon pressing USER1KEY functin key*/
    short   shXu2key;    /* userexit -  Upon pressing USER2KEY functin key*/
    short   shXu3key;    /* userexit -  Upon pressing USER2KEY functin key*/
    short   shXu4key;    /* userexit -  Upon pressing USER2KEY functin key*/
    short   shXu5key;    /* userexit -  Upon pressing USER2KEY functin key*/
    short   shXfctlkey;   /* userexit -  Upon pressing HELPKEY function key*/
    short   shXgctlkey;   /* userexit -  Upon pressing HELPKEY function key*/
    short   shXnctlkey;   /* userexit -  Upon pressing HELPKEY function key*/
    short   shXrctlkey;   /* userexit -  Upon pressing HELPKEY function key*/
    short   shXtctlkey;   /* userexit -  Upon pressing HELPKEY function key*/
    short   shXwctlkey;   /* userexit -  Upon pressing HELPKEY function key*/
    short   shXxctlkey;   /* userexit -  Upon pressing HELPKEY function key*/
    short   shXzctlkey;   /* userexit -  Upon pressing HELPKEY function key*/
    short   shXmousekeylabel; /* userexit -Upon pressing JUMPKEY function key*/
    short   shXmousekey;  /* userexit - Upon pressing JUMPKEY function key*/

    SCRFIELD *shScrfield; /* index into array of structs of type SCRFIELD */
    short   shSfnum;    /* number of fields composing the screen */

    short   shDfbreak;  /* REPORT: index into DBFIELD for page breaks */
    short   shDfbrklen; /* REPORT: significant length of break field */
    short   shSelnum;   /* REPORT: number of select statments */
    short   shSelect;   /* REPORT: select statement */
    short   shAttr;     /* REPORT: */
    short   shFormat;   /* REPORT: */
    short   shFormat2;   /* REPORT: */
    short   shPagelen;  /* REPORT: */
    short   shPrintlen;  /* REPORT: */

    short   shState;    /* state of screen fields - updated by system */
    char   *shBuffer;   /* screen buffer */
    char   *shPrevbuf;  /* screen buffer */
    char   *shOrigbuf;  /* screen buffer */
    char   *shLastbuf;  /* screen buffer */
    KEYTABLE *shKeytbl; /* table of valid keys when entering screen field */
                        /* updated by user  */
    TRANSIENT shTransient;
 } SCRHEAD;



/***********************************************************************
*                                                                      *
*     I N S T R U C T I O N    S T R U C T U R E S                     *
*                                                                      *
************************************************************************/


typedef struct
{
    short   ihName;   /* ---> offset into RESIDENT text array   */
    short   ihType;   /* C=C-function, R=Resident Instructions,
                         T=Transient Instructions */
    short   ihIndex;  /* index into array of structures INSFIELD */
                      /* routine exits upon reaching exit command */
    short   ihDummy;  /* force alignment */
} INSHEAD;


typedef struct
{
    short   ifIndex;   /*  */
    char    ifType;    /* R = screen head,  E= database HEAD
                          S = screen field, D= database field
                          C = Command     , V="C" variable, 
                          I= IF statment
                       */
    char    ifConnect;  /* always '=' for now include 'CONNECTOR' here */
} INSFIELD;


/***********************************************************************
*                                                                      *
*          M A S T E R         S T R U C T U R E S                     *
*                                                                      *
************************************************************************/



/* structure types  */
#define  S_DBHEAD        0
#define  S_SCRHEAD       1
#define  S_INSHEAD       2
#define  S_DBFIELD       3
#define  S_SCRFIELD      4
#define  S_INSFIELD      5
#define  S_DBINDEX       6

#define  S_RESTEXT       7
#define  S_SYMTEXT       8
#define  S_TEMPTEXT      9
#define  S_TRANMAX      10

#define  S_SCRTRANS     11
#define  S_MAXTABLE     10

typedef struct {
    char   *suPtr;
    short   suPtrsize;
    short   suNum;      /* number of structurs */
    short   suTotsize;  /* Total bytes for all structures/text  */
    char  **suNameptr;  /* pointer to text area where logical name stored */
    short   suNameoffs;
    short   suLenoffs;   /* applies to FIELD structures only */
    short   suTypeoffs;  /* applies to FIELD structures only */
    } STRTABLE;



typedef struct
{
    short      hdDhnum;    /*  Number of DBHEAD structs          */
    long       hdDhsize;
    DBHEAD   **hdDhptr;   /* points to first DBHEAD  structure in array */
    short      hdShnum;    /*  Number of SCRHEAD structs          */
    long       hdShsize;
    SCRHEAD  **hdShptr;   /* points to first SCRHEAD structure in array */
    short      hdIhnum;    /*  Number of INSHEAD structs          */
    long       hdIhsize;
    INSHEAD   *hdIhptr;    /* points to first INSHEAD  structure in array */
    short      hdDfnum;    /*  Number of DBFIELD structs          */
    long       hdDfsize;
    DBFIELD   *hdDfptr;    /* points to first DBFIELD  structure in array */
    short      hdSfnum;    /*  Number of SCRFIELD structs          */
    long       hdSfsize;
    SCRFIELD  *hdSfptr;    /* points to first SCRFIELD structure in array */
    short      hdIfnum;    /*  Number of INSFIELD structs          */
    long       hdIfsize;
    INSFIELD  *hdIfptr;    /* points to first INSFIELD  structure in array */
    short      hdDinum;    /*  Number of DBINDEX structs          */
    long       hdDisize;
    DBINDEX   *hdDiptr;    /* points to first DBINDEX  structure in array */
    short      hd_dum1;    /*  Dummy Field    */
    long       hdRtxtsize; /*  Number of bytes of RESIDENT text    */
    char      *hdRtxtptr;   /*  Number of bytes of RESIDENT text    */
    short      hd_dum2;    /*  Dummy Field    */
    long       hdSymsize;  /* size to allocate for symbol table */
    char      *hdSymptr;  /*  points to begining of symbol space */
    short      hd_dum3;    /*  Dummy Field    */
    long       hdTempsize; /*  Number of bytes of Temporary text    */
    char      *hdTempptr;/*  points to begining of Temporary Text space */
    short      hd_dum4;    /*  Dummy Field    */
    long       hdTranmax; /*  maximum size to allocate for transient info */
    char      *hdTranptr; /*  points to begining of transient space */


 /* Miscellaneous  */

    short      hdVersion;
    long       hdRessize;  /* size to allocate for resident structures.
                            by reading this many bytes all resident structures
                            will have been loaded  */
    char      *hdResptr;/*  points to begining of resident  space */
    long       hdTransize; /*  Total size of transient info */

    long      hdSymseek;  /*  start of SYMBOL TABLE ARRAY  */
    long      hdBufsize;  /*  Number of bytes of all Data Buffers  */
    long      hdBufmax;   /* maximum staging buffer size  */
    char     *hdBufptr;   /* buffers */
    char     *hdStageptr;   /* stageing buffer */
    char     *hd2Stageptr;  /* stageing buffer for transaction processing */
    short     hdMainexit; /* flag telling whether main() exists */
    short     hdUserexit; /* Number of C Userexits */
    char      hdTranname[20];/*  Name of Picture file */
    char      hdPicname[20];/*  Name of Picture file */
    int       hdFdpic;      /*  File Descriptor for Picture file */
    short     hdActive;     /* Active header type (see MACROS above)  */
    short     hdDhmax;     /* Maximum number of Dhptr's */
    short     hdShmax;     /* Maximum number of Shptr's */
    char     *hdDflow;    /* Maximum number of Shptr's */
    char     *hdDfhigh;    /* Maximum number of Shptr's */
    char     *hdSflow;    /* Maximum number of Shptr's */
    char     *hdSfhigh;    /* Maximum number of Shptr's */
    short     hdState; 

} HDHEAD;

typedef struct {
    short hfNum;
    long hfSize;
    char *hfPtr;
    } HDFIELD;

#define MAXHDFIELD 15

/***********************************************************************
*                                                                      *
*             G L O B A L     V A R I A B L E S                        *
*                                                                      *
*************************************************************************/

typedef struct {
  char *ufName;
  int (*ufFunc)();
} USERFUNCS;

typedef struct {
  char      *ueName;
  USERFUNCS *ueUserfuncs;
} USEREXITS;

typedef struct {
    char    cwpgname[18];
} CWPARM;

#define NULLSH  ((SCRHEAD *) 0)
#define NULLSF  ((SCRFIELD *) 0)
#define NULLDH  ((DBHEAD *) 0)
#define NULLDI  ((DBINDEX *) 0)
#define NULLDF  ((DBFIELD *) 0)
#define NULLUF  ((USERFUNCS *) 0)
